@@ -2,12 +2,12 @@
  *
  * rewriteManip.c
  *
- * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /home/projects/pgsql/cvsroot/pgsql/src/backend/rewrite/rewriteManip.c,v 1.56 2001/03/22 03:59:44 momjian Exp $
+ *	  $Header: /cvsroot/pgsql-server/src/backend/rewrite/rewriteManip.c,v 1.67 2002/10/20 00:58:55 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -43,14 +43,13 @@ static bool checkExprHasSubLink_walker(Node *node, void *context);
 bool
 checkExprHasAggs(Node *node)
 {
-
 	/*
 	 * If a Query is passed, examine it --- but we will not recurse into
 	 * sub-Queries.
 	 */
 	if (node && IsA(node, Query))
 		return query_tree_walker((Query *) node, checkExprHasAggs_walker,
-								 NULL, false);
+								 NULL, QTW_IGNORE_SUBQUERIES);
 	else
 		return checkExprHasAggs_walker(node, NULL);
 }
@@ -74,14 +73,13 @@ checkExprHasAggs_walker(Node *node, void *context)
 bool
 checkExprHasSubLink(Node *node)
 {
-
 	/*
 	 * If a Query is passed, examine it --- but we will not recurse into
 	 * sub-Queries.
 	 */
 	if (node && IsA(node, Query))
 		return query_tree_walker((Query *) node, checkExprHasSubLink_walker,
-								 NULL, false);
+								 NULL, QTW_IGNORE_SUBQUERIES);
 	else
 		return checkExprHasSubLink_walker(node, NULL);
 }
@@ -103,8 +101,8 @@ checkExprHasSubLink_walker(Node *node, void *context)
  *
  * Find all Var nodes in the given tree with varlevelsup == sublevels_up,
  * and increment their varno fields (rangetable indexes) by 'offset'.
- * The varnoold fields are adjusted similarly.	Also, RangeTblRef nodes
- * in join trees and setOp trees are adjusted.
+ * The varnoold fields are adjusted similarly.	Also, RangeTblRef and
+ * JoinExpr nodes in join trees and setOp trees are adjusted.
  *
  * NOTE: although this has the form of a walker, we cheat and modify the
  * nodes in-place.	The given expression tree should have been copied
@@ -142,6 +140,14 @@ OffsetVarNodes_walker(Node *node, OffsetVarNodes_context *context)
 		/* the subquery itself is visited separately */
 		return false;
 	}
+	if (IsA(node, JoinExpr))
+	{
+		JoinExpr   *j = (JoinExpr *) node;
+
+		if (context->sublevels_up == 0)
+			j->rtindex += context->offset;
+		/* fall through to examine children */
+	}
 	if (IsA(node, Query))
 	{
 		/* Recurse into subselects */
@@ -149,7 +155,7 @@ OffsetVarNodes_walker(Node *node, OffsetVarNodes_context *context)
 
 		context->sublevels_up++;
 		result = query_tree_walker((Query *) node, OffsetVarNodes_walker,
-								   (void *) context, true);
+								   (void *) context, 0);
 		context->sublevels_up--;
 		return result;
 	}
@@ -190,7 +196,7 @@ OffsetVarNodes(Node *node, int offset, int sublevels_up)
 				lfirsti(l) += offset;
 		}
 		query_tree_walker(qry, OffsetVarNodes_walker,
-						  (void *) &context, true);
+						  (void *) &context, 0);
 	}
 	else
 		OffsetVarNodes_walker(node, &context);
@@ -202,7 +208,7 @@ OffsetVarNodes(Node *node, int offset, int sublevels_up)
  * Find all Var nodes in the given tree belonging to a specific relation
  * (identified by sublevels_up and rt_index), and change their varno fields
  * to 'new_index'.	The varnoold fields are changed too.  Also, RangeTblRef
- * nodes in join trees and setOp trees are adjusted.
+ * and JoinExpr nodes in join trees and setOp trees are adjusted.
  *
  * NOTE: although this has the form of a walker, we cheat and modify the
  * nodes in-place.	The given expression tree should have been copied
@@ -243,6 +249,15 @@ ChangeVarNodes_walker(Node *node, ChangeVarNodes_context *context)
 		/* the subquery itself is visited separately */
 		return false;
 	}
+	if (IsA(node, JoinExpr))
+	{
+		JoinExpr   *j = (JoinExpr *) node;
+
+		if (context->sublevels_up == 0 &&
+			j->rtindex == context->rt_index)
+			j->rtindex = context->new_index;
+		/* fall through to examine children */
+	}
 	if (IsA(node, Query))
 	{
 		/* Recurse into subselects */
@@ -250,7 +265,7 @@ ChangeVarNodes_walker(Node *node, ChangeVarNodes_context *context)
 
 		context->sublevels_up++;
 		result = query_tree_walker((Query *) node, ChangeVarNodes_walker,
-								   (void *) context, true);
+								   (void *) context, 0);
 		context->sublevels_up--;
 		return result;
 	}
@@ -295,7 +310,7 @@ ChangeVarNodes(Node *node, int rt_index, int new_index, int sublevels_up)
 			}
 		}
 		query_tree_walker(qry, ChangeVarNodes_walker,
-						  (void *) &context, true);
+						  (void *) &context, 0);
 	}
 	else
 		ChangeVarNodes_walker(node, &context);
@@ -346,7 +361,7 @@ IncrementVarSublevelsUp_walker(Node *node,
 		context->min_sublevels_up++;
 		result = query_tree_walker((Query *) node,
 								   IncrementVarSublevelsUp_walker,
-								   (void *) context, true);
+								   (void *) context, 0);
 		context->min_sublevels_up--;
 		return result;
 	}
@@ -370,7 +385,7 @@ IncrementVarSublevelsUp(Node *node, int delta_sublevels_up,
 	 */
 	if (node && IsA(node, Query))
 		query_tree_walker((Query *) node, IncrementVarSublevelsUp_walker,
-						  (void *) &context, true);
+						  (void *) &context, 0);
 	else
 		IncrementVarSublevelsUp_walker(node, &context);
 }
@@ -412,6 +427,15 @@ rangeTableEntry_used_walker(Node *node,
 		/* the subquery itself is visited separately */
 		return false;
 	}
+	if (IsA(node, JoinExpr))
+	{
+		JoinExpr   *j = (JoinExpr *) node;
+
+		if (j->rtindex == context->rt_index &&
+			context->sublevels_up == 0)
+			return true;
+		/* fall through to examine children */
+	}
 	if (IsA(node, Query))
 	{
 		/* Recurse into subselects */
@@ -419,7 +443,7 @@ rangeTableEntry_used_walker(Node *node,
 
 		context->sublevels_up++;
 		result = query_tree_walker((Query *) node, rangeTableEntry_used_walker,
-								   (void *) context, true);
+								   (void *) context, 0);
 		context->sublevels_up--;
 		return result;
 	}
@@ -442,7 +466,7 @@ rangeTableEntry_used(Node *node, int rt_index, int sublevels_up)
 	 */
 	if (node && IsA(node, Query))
 		return query_tree_walker((Query *) node, rangeTableEntry_used_walker,
-								 (void *) &context, true);
+								 (void *) &context, 0);
 	else
 		return rangeTableEntry_used_walker(node, &context);
 }
@@ -484,7 +508,7 @@ attribute_used_walker(Node *node,
 
 		context->sublevels_up++;
 		result = query_tree_walker((Query *) node, attribute_used_walker,
-								   (void *) context, true);
+								   (void *) context, 0);
 		context->sublevels_up--;
 		return result;
 	}
@@ -508,7 +532,7 @@ attribute_used(Node *node, int rt_index, int attno, int sublevels_up)
 	 */
 	if (node && IsA(node, Query))
 		return query_tree_walker((Query *) node, attribute_used_walker,
-								 (void *) &context, true);
+								 (void *) &context, 0);
 	else
 		return attribute_used_walker(node, &context);
 }
@@ -549,10 +573,10 @@ getInsertSelectQuery(Query *parsetree, Query ***subquery_ptr)
 	 * they've been pushed down to the SELECT.
 	 */
 	if (length(parsetree->rtable) >= 2 &&
-		strcmp(rt_fetch(PRS2_OLD_VARNO, parsetree->rtable)->eref->relname,
-			   "*OLD*") == 0 &&
-		strcmp(rt_fetch(PRS2_NEW_VARNO, parsetree->rtable)->eref->relname,
-			   "*NEW*") == 0)
+	 strcmp(rt_fetch(PRS2_OLD_VARNO, parsetree->rtable)->eref->aliasname,
+			"*OLD*") == 0 &&
+	 strcmp(rt_fetch(PRS2_NEW_VARNO, parsetree->rtable)->eref->aliasname,
+			"*NEW*") == 0)
 		return parsetree;
 	Assert(parsetree->jointree && IsA(parsetree->jointree, FromExpr));
 	if (length(parsetree->jointree->fromlist) != 1)
@@ -565,10 +589,10 @@ getInsertSelectQuery(Query *parsetree, Query ***subquery_ptr)
 		  selectquery->commandType == CMD_SELECT))
 		elog(ERROR, "getInsertSelectQuery: expected to find SELECT subquery");
 	if (length(selectquery->rtable) >= 2 &&
-	 strcmp(rt_fetch(PRS2_OLD_VARNO, selectquery->rtable)->eref->relname,
-			"*OLD*") == 0 &&
-	 strcmp(rt_fetch(PRS2_NEW_VARNO, selectquery->rtable)->eref->relname,
-			"*NEW*") == 0)
+	strcmp(rt_fetch(PRS2_OLD_VARNO, selectquery->rtable)->eref->aliasname,
+		   "*OLD*") == 0 &&
+	strcmp(rt_fetch(PRS2_NEW_VARNO, selectquery->rtable)->eref->aliasname,
+		   "*NEW*") == 0)
 	{
 		if (subquery_ptr)
 			*subquery_ptr = &(selectrte->subquery);
@@ -592,15 +616,21 @@ AddQual(Query *parsetree, Node *qual)
 
 	if (parsetree->commandType == CMD_UTILITY)
 	{
-
 		/*
-		 * Noplace to put the qual on a utility statement.
+		 * There's noplace to put the qual on a utility statement.
 		 *
-		 * For now, we expect utility stmt to be a NOTIFY, so give a specific
-		 * error message for that case.
+		 * If it's a NOTIFY, silently ignore the qual; this means that the
+		 * NOTIFY will execute, whether or not there are any qualifying
+		 * rows. While clearly wrong, this is much more useful than
+		 * refusing to execute the rule at all, and extra NOTIFY events
+		 * are harmless for typical uses of NOTIFY.
+		 *
+		 * If it isn't a NOTIFY, error out, since unconditional execution of
+		 * other utility stmts is unlikely to be wanted.  (This case is
+		 * not currently allowed anyway, but keep the test for safety.)
 		 */
 		if (parsetree->utilityStmt && IsA(parsetree->utilityStmt, NotifyStmt))
-			elog(ERROR, "Conditional NOTIFY is not implemented");
+			return;
 		else
 			elog(ERROR, "Conditional utility statements are not implemented");
 	}
@@ -634,15 +664,13 @@ AddHavingQual(Query *parsetree, Node *havingQual)
 
 	if (parsetree->commandType == CMD_UTILITY)
 	{
-
 		/*
-		 * Noplace to put the qual on a utility statement.
+		 * There's noplace to put the qual on a utility statement.
 		 *
-		 * For now, we expect utility stmt to be a NOTIFY, so give a specific
-		 * error message for that case.
+		 * See comments in AddQual for motivation.
 		 */
 		if (parsetree->utilityStmt && IsA(parsetree->utilityStmt, NotifyStmt))
-			elog(ERROR, "Conditional NOTIFY is not implemented");
+			return;
 		else
 			elog(ERROR, "Conditional utility statements are not implemented");
 	}
@@ -662,35 +690,26 @@ AddHavingQual(Query *parsetree, Node *havingQual)
 	parsetree->hasSubLinks |= checkExprHasSubLink(copy);
 }
 
-#ifdef NOT_USED
+
+/*
+ * Invert the given clause and add it to the WHERE qualifications of the
+ * given querytree.  Inversion means "x IS NOT TRUE", not just "NOT x",
+ * else we will do the wrong thing when x evaluates to NULL.
+ */
 void
-AddNotHavingQual(Query *parsetree, Node *havingQual)
+AddInvertedQual(Query *parsetree, Node *qual)
 {
-	Node	   *notqual;
-
-	if (havingQual == NULL)
-		return;
-
-	/* Need not copy input qual, because AddHavingQual will... */
-	notqual = (Node *) make_notclause((Expr *) havingQual);
-
-	AddHavingQual(parsetree, notqual);
-}
-
-#endif
-
-void
-AddNotQual(Query *parsetree, Node *qual)
-{
-	Node	   *notqual;
+	BooleanTest *invqual;
 
 	if (qual == NULL)
 		return;
 
 	/* Need not copy input qual, because AddQual will... */
-	notqual = (Node *) make_notclause((Expr *) qual);
+	invqual = makeNode(BooleanTest);
+	invqual->arg = qual;
+	invqual->booltesttype = IS_NOT_TRUE;
 
-	AddQual(parsetree, notqual);
+	AddQual(parsetree, (Node *) invqual);
 }
 
 
@@ -729,7 +748,6 @@ FindMatchingTLEntry(List *tlist, char *e_attname)
 	}
 	return NULL;
 }
-
 #endif
 
 
@@ -765,8 +783,13 @@ ResolveNew_mutator(Node *node, ResolveNew_context *context)
 		if (this_varno == context->target_varno &&
 			this_varlevelsup == context->sublevels_up)
 		{
-			Node	   *n = FindMatchingNew(context->targetlist,
-											var->varattno);
+			Node	   *n;
+
+			/* band-aid: don't do the wrong thing with a whole-tuple Var */
+			if (var->varattno == InvalidAttrNumber)
+				elog(ERROR, "ResolveNew: can't handle whole-tuple reference");
+
+			n = FindMatchingNew(context->targetlist, var->varattno);
 
 			if (n == NULL)
 			{
@@ -820,7 +843,7 @@ ResolveNew_mutator(Node *node, ResolveNew_context *context)
 
 		FLATCOPY(newnode, query, Query);
 		context->sublevels_up++;
-		query_tree_mutator(newnode, ResolveNew_mutator, context, true);
+		query_tree_mutator(newnode, ResolveNew_mutator, context, 0);
 		context->sublevels_up--;
 		return (Node *) newnode;
 	}
@@ -852,7 +875,7 @@ ResolveNew(Node *node, int target_varno, int sublevels_up,
 
 		FLATCOPY(newnode, query, Query);
 		query_tree_mutator(newnode, ResolveNew_mutator,
-						   (void *) &context, true);
+						   (void *) &context, 0);
 		return (Node *) newnode;
 	}
 	else
@@ -879,7 +902,7 @@ typedef struct
 	int		   *modified;
 	int		   *badsql;
 	int			sublevels_up;
-}			HandleRIRAttributeRule_context;
+}	HandleRIRAttributeRule_context;
 
 static Node *
 HandleRIRAttributeRule_mutator(Node *node,
@@ -960,7 +983,7 @@ HandleRIRAttributeRule_mutator(Node *node,
 		FLATCOPY(newnode, query, Query);
 		context->sublevels_up++;
 		query_tree_mutator(newnode, HandleRIRAttributeRule_mutator,
-						   context, true);
+						   context, 0);
 		context->sublevels_up--;
 		return (Node *) newnode;
 	}
@@ -988,7 +1011,7 @@ HandleRIRAttributeRule(Query *parsetree,
 	context.sublevels_up = 0;
 
 	query_tree_mutator(parsetree, HandleRIRAttributeRule_mutator,
-					   (void *) &context, true);
+					   (void *) &context, 0);
 }
 
-#endif	 /* NOT_USED */
+#endif   /* NOT_USED */
